import base64
from typing import TYPE_CHECKING, Any, ClassVar

import pydantic
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema

if TYPE_CHECKING:
    from numpy.typing import NDArray


class NumpyArrayField:
    """Wrapper class to support numpy arrays as fields on pydantic models and provide
    JSON or binary serialization implementations.

    The JSON serialization exposes (data, shape, dtype), and the data is base-64
    encoded which leads to ~33% overhead. A more compact serialization can be achieved
    using ``msgpack_numpy`` (integrated in chains, if RPC-option ``use_binary`` is
    enabled).

    Usage example:

    ```
    import numpy as np

    class MyModel(pydantic.BaseModel):
        my_array: NumpyArrayField

    m = MyModel(my_array=np.arange(4).reshape((2, 2)))
    m.my_array.array += 10  # Work with the numpy array.
    print(m)
    # my_array=NumpyArrayField(
    #   shape=(2, 2),
    #   dtype=int64,
    #   data=[[10 11] [12 13]])
    m_json = m.model_dump_json()  # Serialize.
    print(m_json)
    # {"my_array":{"data_b64":"CgAAAAAAAAALAAAAAAAAAAwAAAAAAAAADQAAAAAAAAA=","shape":[2,2],"dtype":"int64"}}
    m2 = MyModel.model_validate_json(m_json)  # De-serialize.
    ```
    """

    data_key: ClassVar[str] = "data_b64"
    shape_key: ClassVar[str] = "shape"
    dtype_key: ClassVar[str] = "dtype"
    array: "NDArray"

    def __init__(self, array: "NDArray"):
        self.array = array

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(shape={self.array.shape}, "
            f"dtype={self.array.dtype}, data={self.array})"
        )

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: pydantic.GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls.validate_numpy_array,
            core_schema.any_schema(),
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls.serialize_numpy_array, info_arg=True
            ),
        )

    @classmethod
    def validate_numpy_array(cls, value: Any):
        import numpy as np

        keys = {cls.data_key, cls.shape_key, cls.dtype_key}
        if isinstance(value, dict) and keys.issubset(value):
            try:
                data = base64.b64decode(value[cls.data_key])
                array = np.frombuffer(data, dtype=value[cls.dtype_key]).reshape(
                    value[cls.shape_key]
                )
                return cls(array)
            except (ValueError, TypeError) as e:
                raise TypeError(
                    "numpy_array_validation"
                    f"Invalid data, shape, or dtype for NumPy array: {str(e)}"
                )
        if isinstance(value, np.ndarray):
            return cls(value)
        if isinstance(value, cls):
            return value

        raise TypeError(
            "numpy_array_validation\n"
            f"Expected a NumPy array or a dictionary with keys {keys}.\n"
            f"Got:\n{value}"
        )

    @classmethod
    def serialize_numpy_array(
        cls, obj: "NumpyArrayField", info: core_schema.SerializationInfo
    ):
        if info.mode == "json":
            return {
                cls.data_key: base64.b64encode(obj.array.tobytes()).decode("utf-8"),
                cls.shape_key: obj.array.shape,
                cls.dtype_key: str(obj.array.dtype),
            }
        return obj.array

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _core_schema: core_schema.CoreSchema,
        handler: pydantic.GetJsonSchemaHandler,
    ) -> JsonSchemaValue:
        json_schema = handler(_core_schema)
        json_schema.update(
            {
                "type": "object",
                "properties": {
                    "data": {"type": "string", "format": "byte"},
                    "shape": {
                        "type": "array",
                        "items": {"type": "integer"},
                        "minItems": 1,
                    },
                    "dtype": {"type": "string"},
                },
                "required": ["data", "shape", "dtype"],
            }
        )
        return json_schema



import logging
import os
import urllib.parse
from typing import Any, Mapping

from pythonjsonlogger import jsonlogger

LOCAL_DATE_FORMAT = "%H:%M:%S"


def _disable_json_logging() -> bool:
    return bool(os.environ.get("DISABLE_JSON_LOGGING"))


class _HealthCheckFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        excluded_paths = {
            "GET / ",
            "GET /v1/models/model ",
            "GET /v1/models/model/loaded ",
        }
        msg = record.getMessage()
        return not any(path in msg for path in excluded_paths)


class _AccessJsonFormatter(jsonlogger.JsonFormatter):
    def format(self, record: logging.LogRecord) -> str:
        # Uvicorn sets record.msg = '%s - "%s %s HTTP/%s" %d' and
        # record.args = (addr, method, path, version, status).
        # Python's logging system resolves final
        # record.message = record.msg % record.args unless we override record.msg.
        if record.name == "uvicorn.access" and record.args and len(record.args) == 5:
            client_addr, method, raw_path, version, status = record.args
            path_decoded = urllib.parse.unquote(str(raw_path))
            new_message = (
                f"Handled request from {client_addr} - {method} "
                f"{path_decoded} HTTP/{version} {status}"
            )
            record.msg = new_message
            record.args = ()  # Ensure Python doesn't reapply the old format string
        return super().format(record)


class _AccessFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        if record.name == "uvicorn.access" and record.args and len(record.args) == 5:
            client_addr, method, raw_path, version, status = record.args
            path_decoded = urllib.parse.unquote(str(raw_path))
            new_message = (
                f"Handled request from  {client_addr} - {method} "
                f"{path_decoded} HTTP/{version} {status}"
            )
            record.msg = new_message
            record.args = ()
        return super().format(record)


def make_log_config(log_level: str) -> Mapping[str, Any]:
    # Warning: `ModelWrapper` depends on correctly setup `uvicorn` logger,
    # if you change/remove that logger, make sure `ModelWrapper` has a suitable
    # alternative logger that is also correctly setup in the load thread.
    formatters = (
        {
            "default_formatter": {
                "format": "%(asctime)s.%(msecs)04d %(levelname)s %(message)s",
                "datefmt": LOCAL_DATE_FORMAT,
            },
            "access_formatter": {
                "()": _AccessFormatter,
                "format": "%(asctime)s.%(msecs)04d %(levelname)s %(message)s",
                "datefmt": LOCAL_DATE_FORMAT,
            },
        }
        if _disable_json_logging()
        else {
            "default_formatter": {
                "()": jsonlogger.JsonFormatter,
                "format": "%(asctime)s %(levelname)s %(message)s",
            },
            "access_formatter": {
                "()": _AccessJsonFormatter,
                "format": "%(asctime)s %(levelname)s %(message)s",
            },
        }
    )

    log_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "filters": {"health_check_filter": {"()": _HealthCheckFilter}},
        "formatters": formatters,
        "handlers": {
            "default_handler": {
                "formatter": "default_formatter",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stderr",
            },
            "access_handler": {
                "formatter": "access_formatter",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
            },
        },
        "loggers": {
            "uvicorn": {
                "handlers": ["default_handler"],
                "level": log_level,
                "propagate": False,
            },
            "uvicorn.error": {
                "handlers": ["default_handler"],
                "level": "INFO",
                "propagate": False,
            },
            "uvicorn.access": {
                "handlers": ["access_handler"],
                "level": "INFO",
                "propagate": False,
                "filters": ["health_check_filter"],
            },
        },
        # Catch-all for module loggers
        "root": {"handlers": ["default_handler"], "level": log_level},
    }
    return log_config



import inspect
from typing import AsyncGenerator, Awaitable, Generator, Union

from pydantic import BaseModel
from truss.templates.server.common.schema import TrussSchema


class ModelInput(BaseModel):
    input: str
    stream: bool


class ModelOutput(BaseModel):
    output: str


def test_truss_schema_pydantic_empty_annotations():
    class Model:
        def predict(self, request):
            return "hello"

    model = Model()

    input_signature = inspect.signature(model.predict).parameters
    output_signature = inspect.signature(model.predict).return_annotation

    schema = TrussSchema.from_signature(input_signature, output_signature)

    assert schema is None


def test_truss_schema_pydantic_input_and_output():
    class Model:
        def predict(self, request: ModelInput) -> ModelOutput:
            return ModelOutput(output=request.input)

    model = Model()

    input_signature = inspect.signature(model.predict).parameters
    output_signature = inspect.signature(model.predict).return_annotation

    schema = TrussSchema.from_signature(input_signature, output_signature)

    assert schema.input_type == ModelInput
    assert schema.output_type == ModelOutput
    assert not schema.supports_streaming


def test_truss_schema_non_pydantic_input():
    class Model:
        def predict(self, request: str) -> ModelOutput:
            return ModelOutput(output=request)

    model = Model()

    input_signature = inspect.signature(model.predict).parameters
    output_signature = inspect.signature(model.predict).return_annotation

    schema = TrussSchema.from_signature(input_signature, output_signature)

    assert schema is None


def test_truss_schema_non_pydantic_output():
    class Model:
        def predict(self, request: ModelInput) -> str:
            return request.input

    model = Model()

    input_signature = inspect.signature(model.predict).parameters
    output_signature = inspect.signature(model.predict).return_annotation

    schema = TrussSchema.from_signature(input_signature, output_signature)

    assert schema is None


def test_truss_schema_list_types():
    class Model:
        def predict(self, request: list[str]) -> list[str]:
            return ["foo", "bar"]

    model = Model()

    input_signature = inspect.signature(model.predict).parameters
    output_signature = inspect.signature(model.predict).return_annotation

    schema = TrussSchema.from_signature(input_signature, output_signature)

    assert schema is None


def test_truss_schema_dict_types():
    class Model:
        def predict(self, request: dict[str, str]) -> dict[str, str]:
            return {"foo": "bar"}

    model = Model()

    input_signature = inspect.signature(model.predict).parameters
    output_signature = inspect.signature(model.predict).return_annotation

    schema = TrussSchema.from_signature(input_signature, output_signature)

    assert schema is None


def test_truss_schema_async():
    class Model:
        async def predict(self, request: ModelInput) -> Awaitable[ModelOutput]:
            return ModelOutput(output=request.input)

    model = Model()

    input_signature = inspect.signature(model.predict).parameters
    output_signature = inspect.signature(model.predict).return_annotation

    schema = TrussSchema.from_signature(input_signature, output_signature)

    assert schema.input_type == ModelInput
    assert schema.output_type == ModelOutput
    assert not schema.supports_streaming


def test_truss_schema_streaming():
    class Model:
        def predict(self, request: ModelInput) -> Generator[str, None, None]:
            yield "hello"

    model = Model()

    input_signature = inspect.signature(model.predict).parameters
    output_signature = inspect.signature(model.predict).return_annotation

    schema = TrussSchema.from_signature(input_signature, output_signature)

    assert schema.input_type == ModelInput
    assert schema.output_type is None
    assert schema.supports_streaming


def test_truss_schema_streaming_async():
    class Model:
        async def predict(self, request: ModelInput) -> AsyncGenerator[str, None]:
            yield "hello"

    model = Model()

    input_signature = inspect.signature(model.predict).parameters
    output_signature = inspect.signature(model.predict).return_annotation

    schema = TrussSchema.from_signature(input_signature, output_signature)

    assert schema.input_type == ModelInput
    assert schema.output_type is None
    assert schema.supports_streaming


def test_truss_schema_union_sync():
    class Model:
        def predict(
            self, request: ModelInput
        ) -> Union[ModelOutput, Generator[str, None, None]]:
            if request.stream:
                return (yield "hello")
            else:
                return ModelOutput(output=request.input)

    model = Model()

    input_signature = inspect.signature(model.predict).parameters
    output_signature = inspect.signature(model.predict).return_annotation

    schema = TrussSchema.from_signature(input_signature, output_signature)
    assert schema.input_type == ModelInput
    assert schema.output_type == ModelOutput
    assert schema.supports_streaming


def test_truss_schema_union_async():
    class Model:
        async def predict(
            self, request: ModelInput
        ) -> Union[Awaitable[ModelOutput], AsyncGenerator[str, None]]:
            if request.stream:

                def inner():
                    for i in range(2):
                        yield str(i)

                return inner()

            return ModelOutput(output=request.input)

    model = Model()

    input_signature = inspect.signature(model.predict).parameters
    output_signature = inspect.signature(model.predict).return_annotation

    schema = TrussSchema.from_signature(input_signature, output_signature)
    assert schema.input_type == ModelInput
    assert schema.output_type is ModelOutput
    assert schema.supports_streaming


def test_truss_schema_union_async_non_pydantic():
    class Model:
        async def predict(
            self, request: ModelInput
        ) -> Union[Awaitable[str], AsyncGenerator[str, None]]:
            return "hello"

    model = Model()

    input_signature = inspect.signature(model.predict).parameters
    output_signature = inspect.signature(model.predict).return_annotation

    schema = TrussSchema.from_signature(input_signature, output_signature)
    assert schema is None


def test_truss_schema_union_non_pydantic():
    class Model:
        def predict(self, request: ModelInput) -> Union[str, int]:
            return "hello"

    model = Model()

    input_signature = inspect.signature(model.predict).parameters
    output_signature = inspect.signature(model.predict).return_annotation

    schema = TrussSchema.from_signature(input_signature, output_signature)

    assert schema is None


def test_truss_schema_async_non_pydantic():
    class Model:
        async def predict(self, request: str) -> Awaitable[str]:
            return "hello"

    model = Model()

    input_signature = inspect.signature(model.predict).parameters
    output_signature = inspect.signature(model.predict).return_annotation

    schema = TrussSchema.from_signature(input_signature, output_signature)
    assert schema is None


def test_truss_schema_union_three_arms():
    class ModelOutput2(BaseModel):
        output2: str

    class ModelOutput3(BaseModel):
        output3: str

    class Model:
        def predict(
            self, request: ModelInput
        ) -> Union[ModelOutput, ModelOutput2, ModelOutput3]:
            return ModelOutput(output=request.input)

    model = Model()
    input_signature = inspect.signature(model.predict).parameters
    output_signature = inspect.signature(model.predict).return_annotation

    schema = TrussSchema.from_signature(input_signature, output_signature)

    assert schema is None



from abc import ABC, abstractmethod
from pathlib import Path


class TrussContext(ABC):
    """Marker class for Truss context.

    A model is represented in a standard form called Truss. Truss contexts
    perform a certain operation on this Truss. Some examples are: running the
    model directly, building an image and, deploying to baseten.
    """

    @staticmethod
    @abstractmethod
    def run(truss_dir: Path):
        pass



import json
import pathlib
import time
from typing import Any, Generator, Sequence

import opentelemetry.sdk.resources as resources
import opentelemetry.sdk.trace as sdk_trace
import opentelemetry.sdk.trace.export as trace_export
from opentelemetry import trace
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import SimpleSpanProcessor


class JSONFileExporter(trace_export.SpanExporter):
    # Copied form truss/templates/server/common/tracing.py.

    def __init__(self, file_path: pathlib.Path):
        self._file = file_path.open("a")

    def export(
        self, spans: Sequence[sdk_trace.ReadableSpan]
    ) -> trace_export.SpanExportResult:
        for span in spans:
            # Get rid of newlines and whitespace.
            self._file.write(json.dumps(json.loads(span.to_json())))
            self._file.write("\n")
        self._file.flush()
        return trace_export.SpanExportResult.SUCCESS

    def shutdown(self) -> None:
        self._file.close()


trace.set_tracer_provider(
    TracerProvider(resource=Resource.create({resources.SERVICE_NAME: "UserModel"}))
)
tracer = trace.get_tracer(__name__)
file_exporter = JSONFileExporter(pathlib.Path("/tmp/otel_user_traces.ndjson"))
span_processor = SimpleSpanProcessor(file_exporter)
trace.get_tracer_provider().add_span_processor(span_processor)


class Model:
    def __init__(self, **kwargs) -> None:
        self._data_dir = kwargs["data_dir"]
        self._config = kwargs["config"]
        self._secrets = kwargs["secrets"]
        self._model = None

    @tracer.start_as_current_span("load_model")
    def load(self):
        pass

    @tracer.start_as_current_span("predict")
    def predict(self, model_input: Any) -> Generator[str, None, None]:
        with tracer.start_as_current_span("start-predict") as span:

            def inner():
                time.sleep(0.02)
                for i in range(5):
                    span.add_event("yield")
                    yield str(i)

            return inner()



from urllib import parse

import pytest
import requests_mock
import truss
from truss.remote.baseten.core import (
    ModelId,
    ModelName,
    ModelVersionId,
    create_chain_atomic,
)
from truss.remote.baseten.custom_types import ChainletDataAtomic, OracleData
from truss.remote.baseten.error import RemoteError
from truss.remote.baseten.remote import BasetenRemote
from truss.truss_handle.truss_handle import TrussHandle

_TEST_REMOTE_URL = "http://test_remote.com"
_TEST_REMOTE_GRAPHQL_PATH = "http://test_remote.com/graphql/"


def assert_request_matches_expected_query(request, expected_query) -> None:
    unescaped_content = parse.unquote_plus(request.text)
    actual_lines = tuple(
        line.strip()
        for line in unescaped_content.replace("query=", "").strip().split("\n")
        if line.strip()
    )
    expected_lines = tuple(
        line.strip() for line in expected_query.split("\n") if line.strip()
    )
    assert actual_lines == expected_lines


def test_get_service_by_version_id():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")

    version = {"id": "version_id", "oracle": {"id": "model_id"}}
    model_version_response = {"data": {"model_version": version}}

    with requests_mock.Mocker() as m:
        m.post(_TEST_REMOTE_GRAPHQL_PATH, json=model_version_response)
        service = remote.get_service(model_identifier=ModelVersionId("version_id"))

    assert service.model_id == "model_id"
    assert service.model_version_id == "version_id"


def test_get_service_by_version_id_no_version():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")
    model_version_response = {"errors": [{"message": "error"}]}
    with requests_mock.Mocker() as m:
        m.post(_TEST_REMOTE_GRAPHQL_PATH, json=model_version_response)
        with pytest.raises(RemoteError):
            remote.get_service(model_identifier=ModelVersionId("version_id"))


def test_get_service_by_model_name():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")

    versions = [
        {"id": "1", "is_draft": False, "is_primary": False},
        {"id": "2", "is_draft": False, "is_primary": True},
        {"id": "3", "is_draft": True, "is_primary": False},
    ]
    model_response = {
        "data": {
            "model": {"name": "model_name", "id": "model_id", "versions": versions}
        }
    }

    with requests_mock.Mocker() as m:
        m.post(_TEST_REMOTE_GRAPHQL_PATH, json=model_response)

        # Check that the production version is returned when published is True.
        service = remote.get_service(
            model_identifier=ModelName("model_name"), published=True
        )
        assert service.model_id == "model_id"
        assert service.model_version_id == "2"

        # Check that the development version is returned when published is False.
        service = remote.get_service(
            model_identifier=ModelName("model_name"), published=False
        )
        assert service.model_id == "model_id"
        assert service.model_version_id == "3"


def test_get_service_by_model_name_no_dev_version():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")

    versions = [{"id": "1", "is_draft": False, "is_primary": True}]
    model_response = {
        "data": {
            "model": {"name": "model_name", "id": "model_id", "versions": versions}
        }
    }

    with requests_mock.Mocker() as m:
        m.post(_TEST_REMOTE_GRAPHQL_PATH, json=model_response)

        # Check that the production version is returned when published is True.
        service = remote.get_service(
            model_identifier=ModelName("model_name"), published=True
        )
        assert service.model_id == "model_id"
        assert service.model_version_id == "1"

        # Since no development version exists, calling get_service with
        # published=False should raise an error.
        with pytest.raises(RemoteError):
            remote.get_service(
                model_identifier=ModelName("model_name"), published=False
            )


def test_get_service_by_model_name_no_prod_version():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")

    versions = [{"id": "1", "is_draft": True, "is_primary": False}]
    model_response = {
        "data": {
            "model": {"name": "model_name", "id": "model_id", "versions": versions}
        }
    }

    with requests_mock.Mocker() as m:
        m.post(_TEST_REMOTE_GRAPHQL_PATH, json=model_response)

        # Since no production version exists, calling get_service with
        # published=True should raise an error.
        with pytest.raises(RemoteError):
            remote.get_service(model_identifier=ModelName("model_name"), published=True)

        # Check that the development version is returned when published is False.
        service = remote.get_service(
            model_identifier=ModelName("model_name"), published=False
        )
        assert service.model_id == "model_id"
        assert service.model_version_id == "1"


def test_get_service_by_model_id():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")

    model_response = {
        "data": {
            "model": {
                "name": "model_name",
                "id": "model_id",
                "primary_version": {"id": "version_id"},
            }
        }
    }

    with requests_mock.Mocker() as m:
        m.post(_TEST_REMOTE_GRAPHQL_PATH, json=model_response)

        service = remote.get_service(model_identifier=ModelId("model_id"))
        assert service.model_id == "model_id"
        assert service.model_version_id == "version_id"


def test_get_service_by_model_id_no_model():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")
    model_response = {"errors": [{"message": "error"}]}
    with requests_mock.Mocker() as m:
        m.post(_TEST_REMOTE_GRAPHQL_PATH, json=model_response)
        with pytest.raises(RemoteError):
            remote.get_service(model_identifier=ModelId("model_id"))


def test_push_raised_value_error_when_deployment_name_and_not_publish(
    custom_model_truss_dir_with_pre_and_post,
):
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")
    model_response = {
        "data": {
            "model": {
                "name": "model_name",
                "id": "model_id",
                "primary_version": {"id": "version_id"},
            }
        }
    }
    with requests_mock.Mocker() as m:
        m.post(_TEST_REMOTE_GRAPHQL_PATH, json=model_response)
        th = TrussHandle(custom_model_truss_dir_with_pre_and_post)

        with pytest.raises(
            ValueError,
            match="Deployment name cannot be used for development deployment",
        ):
            remote.push(
                th,
                "model_name",
                publish=False,
                trusted=False,
                promote=False,
                preserve_previous_prod_deployment=False,
                deployment_name="dep_name",
            )


def test_push_raised_value_error_when_deployment_name_is_not_valid(
    custom_model_truss_dir_with_pre_and_post,
):
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")
    model_response = {
        "data": {
            "model": {
                "name": "model_name",
                "id": "model_id",
                "primary_version": {"id": "version_id"},
            }
        }
    }
    with requests_mock.Mocker() as m:
        m.post(_TEST_REMOTE_GRAPHQL_PATH, json=model_response)
        th = TrussHandle(custom_model_truss_dir_with_pre_and_post)

        with pytest.raises(
            ValueError,
            match="Deployment name must only contain alphanumeric, -, _ and . characters",
        ):
            remote.push(
                th,
                "model_name",
                publish=True,
                trusted=False,
                promote=False,
                preserve_previous_prod_deployment=False,
                deployment_name="dep//name",
            )


def test_push_raised_value_error_when_keep_previous_prod_settings_and_not_promote(
    custom_model_truss_dir_with_pre_and_post,
):
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")
    model_response = {
        "data": {
            "model": {
                "name": "model_name",
                "id": "model_id",
                "primary_version": {"id": "version_id"},
            }
        }
    }
    with requests_mock.Mocker() as m:
        m.post(_TEST_REMOTE_GRAPHQL_PATH, json=model_response)
        th = TrussHandle(custom_model_truss_dir_with_pre_and_post)

        with pytest.raises(
            ValueError,
            match="preserve-previous-production-deployment can only be used with the '--promote' option",
        ):
            remote.push(
                th,
                "model_name",
                publish=False,
                trusted=False,
                promote=False,
                preserve_previous_prod_deployment=True,
            )


def test_create_chain_with_no_publish():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")

    with requests_mock.Mocker() as m:
        m.post(
            _TEST_REMOTE_GRAPHQL_PATH,
            [
                {"json": {"data": {"chains": []}}},
                {
                    "json": {
                        "data": {
                            "deploy_chain_atomic": {
                                "chain_id": "new-chain-id",
                                "chain_deployment_id": "new-chain-deployment-id",
                                "entrypoint_model_id": "new-entrypoint-model-id",
                                "entrypoint_model_version_id": "new-entrypoint-model-version-id",
                            }
                        }
                    }
                },
            ],
        )

        deployment_handle = create_chain_atomic(
            api=remote.api,
            chain_name="draft_chain",
            entrypoint=ChainletDataAtomic(
                name="chainlet-1",
                oracle=OracleData(
                    model_name="model-1",
                    s3_key="s3-key-1",
                    encoded_config_str="encoded-config-str-1",
                    is_trusted=True,
                ),
            ),
            dependencies=[],
            is_draft=True,
            environment=None,
        )

        get_chains_graphql_request = m.request_history[0]
        create_chain_graphql_request = m.request_history[1]

        expected_get_chains_query = """
            {
                chains {
                    id
                    name
                }
            }
        """.strip()

        assert_request_matches_expected_query(
            get_chains_graphql_request, expected_get_chains_query
        )

        chainlets_string = """
            {
                name: "chainlet-1",
                oracle: {
                    model_name: "model-1",
                    s3_key: "s3-key-1",
                    encoded_config_str: "encoded-config-str-1",
                    is_trusted: true,
                    semver_bump: "MINOR"
                }
            }
        """.strip()

        # Note that if publish=False and promote=True, we set publish to True and create
        # a non-draft deployment
        expected_create_chain_mutation = f"""
            mutation {{
                deploy_chain_atomic(
                    chain_name: "draft_chain"
                    is_draft: true
                    entrypoint: {chainlets_string}
                    dependencies: []
                    client_version: "{truss.version()}"
                ) {{
                    chain_id
                    chain_deployment_id
                    entrypoint_model_id
                    entrypoint_model_version_id
                }}
            }}
        """.strip()

        assert_request_matches_expected_query(
            create_chain_graphql_request, expected_create_chain_mutation
        )

        assert deployment_handle.chain_id == "new-chain-id"
        assert deployment_handle.chain_deployment_id == "new-chain-deployment-id"


def test_create_chain_no_existing_chain():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")

    with requests_mock.Mocker() as m:
        m.post(
            _TEST_REMOTE_GRAPHQL_PATH,
            [
                {"json": {"data": {"chains": []}}},
                {
                    "json": {
                        "data": {
                            "deploy_chain_atomic": {
                                "chain_id": "new-chain-id",
                                "chain_deployment_id": "new-chain-deployment-id",
                                "entrypoint_model_id": "new-entrypoint-model-id",
                                "entrypoint_model_version_id": "new-entrypoint-model-version-id",
                            }
                        }
                    }
                },
            ],
        )

        deployment_handle = create_chain_atomic(
            api=remote.api,
            chain_name="new_chain",
            entrypoint=ChainletDataAtomic(
                name="chainlet-1",
                oracle=OracleData(
                    model_name="model-1",
                    s3_key="s3-key-1",
                    encoded_config_str="encoded-config-str-1",
                    is_trusted=True,
                ),
            ),
            dependencies=[],
            is_draft=False,
            environment=None,
        )

        get_chains_graphql_request = m.request_history[0]
        create_chain_graphql_request = m.request_history[1]

        expected_get_chains_query = """
            {
                chains {
                    id
                    name
                }
            }
        """.strip()

        assert_request_matches_expected_query(
            get_chains_graphql_request, expected_get_chains_query
        )

        chainlets_string = """
            {
                name: "chainlet-1",
                oracle: {
                    model_name: "model-1",
                    s3_key: "s3-key-1",
                    encoded_config_str: "encoded-config-str-1",
                    is_trusted: true,
                    semver_bump: "MINOR"
                }
            }
        """.strip()

        expected_create_chain_mutation = f"""
            mutation {{
                deploy_chain_atomic(
                    chain_name: "new_chain"
                    is_draft: false
                    entrypoint: {chainlets_string}
                    dependencies: []
                    client_version: "{truss.version()}"
                ) {{
                    chain_id
                    chain_deployment_id
                    entrypoint_model_id
                    entrypoint_model_version_id
                }}
            }}
        """.strip()

        assert_request_matches_expected_query(
            create_chain_graphql_request, expected_create_chain_mutation
        )

        assert deployment_handle.chain_id == "new-chain-id"
        assert deployment_handle.chain_deployment_id == "new-chain-deployment-id"


def test_create_chain_with_existing_chain_promote_to_environment_publish_false():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")

    with requests_mock.Mocker() as m:
        m.post(
            _TEST_REMOTE_GRAPHQL_PATH,
            [
                {
                    "json": {
                        "data": {
                            "chains": [{"id": "old-chain-id", "name": "old_chain"}]
                        }
                    }
                },
                {
                    "json": {
                        "data": {
                            "deploy_chain_atomic": {
                                "chain_id": "new-chain-id",
                                "chain_deployment_id": "new-chain-deployment-id",
                                "entrypoint_model_id": "new-entrypoint-model-id",
                                "entrypoint_model_version_id": "new-entrypoint-model-version-id",
                            }
                        }
                    }
                },
            ],
        )

        deployment_handle = create_chain_atomic(
            api=remote.api,
            chain_name="old_chain",
            entrypoint=ChainletDataAtomic(
                name="chainlet-1",
                oracle=OracleData(
                    model_name="model-1",
                    s3_key="s3-key-1",
                    encoded_config_str="encoded-config-str-1",
                    is_trusted=True,
                ),
            ),
            dependencies=[],
            is_draft=True,
            environment="production",
        )

        get_chains_graphql_request = m.request_history[0]
        create_chain_graphql_request = m.request_history[1]

        expected_get_chains_query = """
            {
                chains {
                    id
                    name
                }
            }
        """.strip()

        assert_request_matches_expected_query(
            get_chains_graphql_request, expected_get_chains_query
        )

        # Note that if publish=False and environment!=None, we set publish to True and create
        # a non-draft deployment
        chainlets_string = """
            {
                name: "chainlet-1",
                oracle: {
                    model_name: "model-1",
                    s3_key: "s3-key-1",
                    encoded_config_str: "encoded-config-str-1",
                    is_trusted: true,
                    semver_bump: "MINOR"
                }
            }
        """.strip()

        expected_create_chain_mutation = f"""
            mutation {{
                deploy_chain_atomic(
                    chain_id: "old-chain-id"
                    environment: "production"
                    is_draft: false
                    entrypoint: {chainlets_string}
                    dependencies: []
                    client_version: "{truss.version()}"
                ) {{
                    chain_id
                    chain_deployment_id
                    entrypoint_model_id
                    entrypoint_model_version_id
                }}
            }}
        """.strip()

        assert_request_matches_expected_query(
            create_chain_graphql_request, expected_create_chain_mutation
        )

        assert deployment_handle.chain_id == "new-chain-id"
        assert deployment_handle.chain_deployment_id == "new-chain-deployment-id"


def test_create_chain_existing_chain_publish_true_no_promotion():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")

    with requests_mock.Mocker() as m:
        m.post(
            _TEST_REMOTE_GRAPHQL_PATH,
            [
                {
                    "json": {
                        "data": {
                            "chains": [{"id": "old-chain-id", "name": "old_chain"}]
                        }
                    }
                },
                {
                    "json": {
                        "data": {
                            "deploy_chain_atomic": {
                                "chain_id": "new-chain-id",
                                "chain_deployment_id": "new-chain-deployment-id",
                                "entrypoint_model_id": "new-entrypoint-model-id",
                                "entrypoint_model_version_id": "new-entrypoint-model-version-id",
                            }
                        }
                    }
                },
            ],
        )

        deployment_handle = create_chain_atomic(
            api=remote.api,
            chain_name="old_chain",
            entrypoint=ChainletDataAtomic(
                name="chainlet-1",
                oracle=OracleData(
                    model_name="model-1",
                    s3_key="s3-key-1",
                    encoded_config_str="encoded-config-str-1",
                    is_trusted=True,
                ),
            ),
            dependencies=[],
            is_draft=False,
            environment=None,
        )

        get_chains_graphql_request = m.request_history[0]
        create_chain_graphql_request = m.request_history[1]

        expected_get_chains_query = """
            {
                chains {
                    id
                    name
                }
            }
        """.strip()

        assert_request_matches_expected_query(
            get_chains_graphql_request, expected_get_chains_query
        )

        chainlets_string = """
            {
                name: "chainlet-1",
                oracle: {
                    model_name: "model-1",
                    s3_key: "s3-key-1",
                    encoded_config_str: "encoded-config-str-1",
                    is_trusted: true,
                    semver_bump: "MINOR"
                }
            }
        """.strip()

        expected_create_chain_mutation = f"""
            mutation {{
                deploy_chain_atomic(
                    chain_id: "old-chain-id"
                    is_draft: false
                    entrypoint: {chainlets_string}
                    dependencies: []
                    client_version: "{truss.version()}"
                ) {{
                    chain_id
                    chain_deployment_id
                    entrypoint_model_id
                    entrypoint_model_version_id
                }}
            }}
        """.strip()

        assert_request_matches_expected_query(
            create_chain_graphql_request, expected_create_chain_mutation
        )

        assert deployment_handle.chain_id == "new-chain-id"
        assert deployment_handle.chain_deployment_id == "new-chain-deployment-id"


@pytest.mark.parametrize("publish", [True, False])
def test_push_raised_value_error_when_disable_truss_download_for_existing_model(
    publish, custom_model_truss_dir_with_pre_and_post
):
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")
    model_response = {
        "data": {
            "model": {
                "name": "model_name",
                "id": "model_id",
                "primary_version": {"id": "version_id"},
            }
        }
    }
    with requests_mock.Mocker() as m:
        m.post(_TEST_REMOTE_GRAPHQL_PATH, json=model_response)
        th = TrussHandle(custom_model_truss_dir_with_pre_and_post)

        with pytest.raises(
            ValueError, match="disable-truss-download can only be used for new models"
        ):
            remote.push(th, "model_name", publish=publish, disable_truss_download=True)



import json

import aiofiles
import pytest
from truss.templates.shared import dynamic_config_resolver

from truss_chains import definitions


@pytest.mark.parametrize(
    "config",
    [
        {
            "RandInt": {
                "predict_url": "https://model-id.api.baseten.co/deployment/deployment-id/predict"
            }
        },
        {},
        "",
    ],
)
def test_get_dynamic_chainlet_config_value_sync(
    config, tmp_path, dynamic_config_mount_dir
):
    with (tmp_path / definitions.DYNAMIC_CHAINLET_CONFIG_KEY).open("w") as f:
        f.write(json.dumps(config))
    chainlet_service_config = dynamic_config_resolver.get_dynamic_config_value_sync(
        definitions.DYNAMIC_CHAINLET_CONFIG_KEY
    )
    assert json.loads(chainlet_service_config) == config


@pytest.mark.parametrize(
    "config", [{"environment_name": "production", "foo": "bar"}, {}, "", None]
)
def test_get_dynamic_config_environment_value_sync(
    config, tmp_path, dynamic_config_mount_dir
):
    with (tmp_path / dynamic_config_resolver.ENVIRONMENT_DYNAMIC_CONFIG_KEY).open(
        "w"
    ) as f:
        f.write(json.dumps(config))
    environment_str = dynamic_config_resolver.get_dynamic_config_value_sync(
        dynamic_config_resolver.ENVIRONMENT_DYNAMIC_CONFIG_KEY
    )
    assert json.loads(environment_str) == config


def test_get_missing_config_value_sync(dynamic_config_mount_dir):
    chainlet_service_config = dynamic_config_resolver.get_dynamic_config_value_sync(
        definitions.DYNAMIC_CHAINLET_CONFIG_KEY
    )
    assert not chainlet_service_config


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "config",
    [
        {
            "RandInt": {
                "predict_url": "https://model-id.api.baseten.co/deployment/deployment-id/predict"
            }
        },
        {},
        "",
    ],
)
async def test_get_dynamic_chainlet_config_value_async(
    config, tmp_path, dynamic_config_mount_dir
):
    async with aiofiles.open(
        tmp_path / definitions.DYNAMIC_CHAINLET_CONFIG_KEY, "w"
    ) as f:
        await f.write(json.dumps(config))
    chainlet_service_config = (
        await dynamic_config_resolver.get_dynamic_config_value_async(
            definitions.DYNAMIC_CHAINLET_CONFIG_KEY
        )
    )
    assert json.loads(chainlet_service_config) == config


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "config", [{"environment_name": "production", "foo": "bar"}, {}, "", None]
)
async def test_get_dynamic_config_environment_value_async(
    config, tmp_path, dynamic_config_mount_dir
):
    async with aiofiles.open(
        tmp_path / dynamic_config_resolver.ENVIRONMENT_DYNAMIC_CONFIG_KEY, "w"
    ) as f:
        await f.write(json.dumps(config))
    environment_str = await dynamic_config_resolver.get_dynamic_config_value_async(
        dynamic_config_resolver.ENVIRONMENT_DYNAMIC_CONFIG_KEY
    )
    assert json.loads(environment_str) == config


@pytest.mark.asyncio
async def test_get_missing_config_value_async(dynamic_config_mount_dir):
    chainlet_service_config = (
        await dynamic_config_resolver.get_dynamic_config_value_async(
            definitions.DYNAMIC_CHAINLET_CONFIG_KEY
        )
    )
    assert not chainlet_service_config
